import pygraphviz as pgv


def test_name_unicode():
    A = pgv.AGraph(name="unicode")
    assert A.name == "unicode"


def test_node_encoding():
    A = pgv.AGraph(encoding="UTF-8")
    hello = "Здравствуйте!"
    A.add_node(hello)
    n = A.get_node(hello)
    assert n.name == hello

    n.attr["goodbye"] = "До свидания"
    assert n.attr["goodbye"] == "До свидания"


def test_edge_encoding():
    A = pgv.AGraph(encoding="UTF-8")
    hello = "שלום"
    A.add_edge(hello, hello, key=1)  # self loop
    e = A.get_edge(hello, hello)
    assert e.name == "1"
    assert e == (hello, hello)

    e.attr["hello"] = hello
    assert e.attr["hello"] == hello


def test_from_string():
    # test unicode in from_string()
    t = "测试"
    G = pgv.AGraph()
    G.add_node(t)
    ug = str(G)
    sg = str(G)
    G1 = pgv.AGraph(ug)
    G2 = pgv.AGraph(sg)
    assert str(G1) == str(G2)



import pytest

import pygraphviz as pgv


def test_graphviz_version():
    """Test to ensure package exports the __graphviz_version__ attribute, which
    holds the version info for the packaged graphviz version."""
    # Smoke test - the following lines will fail if e.g. __graphviz_version__
    # is not present or the version string takes an unexpected form
    graphviz_version = pgv.__graphviz_version__
    vmaj, vmin, vpatch = graphviz_version.split(".")
    assert int(vmaj) >= 2



import warnings
import pytest
import unittest
import pygraphviz as pgv

stringify = pgv.testing.stringify


def test_agraph_multiedges():
    """See gh-162"""
    G = pgv.AGraph(strict=False)
    G.add_edge("a", "b", "first")
    G.add_edge("a", "b", "second")
    assert sorted(G.edges(keys=True)) == [("a", "b", "first"), ("a", "b", "second")]


def test_tred():
    A = pgv.AGraph(directed=True)
    A.add_edges_from([(0, 1), (1, 2), (0, 2)])
    # copy=False by default
    B = A.tred()
    assert A.edges() == B.edges() == [("0", "1"), ("1", "2")]


def test_tred_copy():
    A = pgv.AGraph(directed=True)
    A.add_edges_from([(0, 1), (1, 2), (0, 2)])
    # With copy=True, original graph should remain unchanged
    B = A.tred(copy=True)
    # NOTE: edges may be reordered, but should remain unaltered
    assert set(A.edges()) == {("0", "1"), ("1", "2"), ("0", "2")}
    assert B.edges() == [("0", "1"), ("1", "2")]


def test_tred_undirected():
    A = pgv.AGraph()
    A.add_edges_from([(0, 1), (1, 2), (0, 2)])
    with pytest.raises(TypeError, match="tred requires a directed graph"):
        A.tred()


def test_tred_cycle():
    """RuntimeWarning raised when cycle detected."""
    A = pgv.AGraph(directed=True)
    A.add_edges_from([(0, 1), (1, 2), (2, 0)])  # Directed 3-cycle
    with warnings.catch_warnings(record=True) as rec:
        A.tred()
    # No transitive reduction possible on 3-cycle
    assert A.edges() == [("0", "1"), ("1", "2"), ("2", "0")]
    assert len(rec) == 1  # Expect 1 RuntimeWarning
    assert rec[0].category == RuntimeWarning
    assert "transitive reduction not unique" in str(rec[0].message)


class TestGraph(unittest.TestCase):
    def setUp(self):
        self.null = pgv.AGraph()
        self.K1 = pgv.AGraph()
        self.K1.add_node(1)
        self.K3 = pgv.AGraph()
        self.K3.add_edges_from([(1, 2), (1, 3), (2, 3)])
        self.K5 = pgv.AGraph()
        self.K5.add_edges_from(
            [
                (1, 2),
                (1, 3),
                (1, 4),
                (1, 5),
                (2, 3),
                (2, 4),
                (2, 5),
                (3, 4),
                (3, 5),
                (4, 5),
            ]
        )
        self.P3 = pgv.AGraph()
        self.P3.add_path([1, 2, 3])
        self.P5 = pgv.AGraph()
        self.P5.add_path([1, 2, 3, 4, 5])

    def test_strict(self):
        A = pgv.AGraph()  # empty
        assert A.is_strict()
        assert A.nodes() == []

    def test_data(self):
        d = {"a": "b", "b": "c"}
        A = pgv.AGraph(data=d)  # with data
        assert sorted(A.nodes()) == ["a", "b", "c"]
        assert sorted(A.edges()) == [("a", "b"), ("b", "c")]

    def test_str(self):
        A = pgv.AGraph(name="test")
        assert stringify(A) == "strict graph test { }"
        A = pgv.AGraph()
        assert stringify(A) == "strict graph { }"

    def test_repr(self):
        A = pgv.AGraph()
        assert repr(A).startswith("<AGraph <Swig Object of type 'Agraph_t")
        # If graph has a name, it should show up in the repr
        A = pgv.AGraph(name="foo")
        assert "foo" in repr(A)

    def test_equal(self):
        A = pgv.AGraph()
        H = pgv.AGraph()
        assert H == A
        assert H is not A

        assert self.P3 == self.P3

        A = pgv.AGraph()
        A.add_path([1, 2, 3])
        assert A.nodes() == self.P3.nodes()
        assert A.edges() == self.P3.edges()
        assert stringify(A) == stringify(self.P3)
        assert A == self.P3

    def test_not_equal(self):
        A = pgv.AGraph()
        A.add_path([1, 2, 3])
        B = self.P3
        assert A == B

        A.add_node(4)
        assert A != B
        A.remove_node(4)
        assert A == B
        A.add_edge(3, 1)
        assert A != B
        A.remove_edge(3, 1)
        assert A == B

        A.add_node(4, hi=9)
        assert A != B
        A.remove_node(4)
        assert A != B  # attribute 'hi' exists for every node.
        B.node_attr["hi"] = ""
        assert A == B
        A.add_edge(3, 1, hi=9)
        assert A != B
        A.remove_edge(3, 1)
        assert A != B
        B.edge_attr["hi"] = ""
        assert A == B

        # Note: adding node attr gives every node that attribute but empty.
        # Default values are only assigned to nodes added after default is set.
        A.node_attr["low"] = 3
        assert A != B
        B.node_attr["low"] = 3
        assert A == B
        B.node_attr["low"] = 4
        assert A == B

        A.edge_attr["low"] = 3
        assert A != B
        B.edge_attr["low"] = 3
        assert A == B
        B.edge_attr["low"] = 4
        assert A == B

        # graph_attr are not default values -- they are attribute values.
        A.graph_attr.update({"low": 5})
        assert A == B
        # print(sorted(A.nodes()), sorted(B.nodes()))
        # print(sorted(A.edges()), sorted(B.edges()))
        # print(tuple(dict(n.attr) for n in sorted(A.nodes())))
        # print(tuple(dict(n.attr) for n in sorted(B.nodes())))
        # print(tuple(dict(e.attr) for e in sorted(A.edges())))
        # print(tuple(dict(e.attr) for e in sorted(B.edges())))
        # print(dict(A.node_attr), dict(B.node_attr))
        # print(dict(A.edge_attr), dict(B.edge_attr))
        # print(dict(A.graph_attr), dict(B.graph_attr))

    def test_hash(self):
        A = pgv.AGraph()
        A.add_path([1, 2, 3])
        assert hash(A) == hash(self.P3)
        B = A.copy()
        assert hash(A) == hash(B)
        B.add_node(4)
        assert hash(A) != hash(B)

    def test_iter(self):
        assert sorted(self.P3.__iter__()) == ["1", "2", "3"]
        assert sorted(self.P3) == ["1", "2", "3"]

    def test_contains(self):
        assert "1" in self.P3
        assert 10 not in self.P3

    def test_len(self):
        assert len(self.P3) == 3

    def test_getitem(self):
        assert self.P3[1] == ["2"]

    def test_missing_getitem(self):
        with pytest.raises(KeyError):
            a = self.P3[10]

    def test_add_remove_node(self):
        A = pgv.AGraph()
        A.add_node(1)
        assert A.nodes() == ["1"]
        A.add_node("A")
        assert sorted(A.nodes()) == ["1", "A"]
        A.add_node([1])  # nodes must be strings or have a __str__
        assert sorted(A.nodes()) == ["1", "A", "[1]"]

        A.remove_node([1])
        assert sorted(A.nodes()) == ["1", "A"]
        A.remove_node("A")
        assert A.nodes() == ["1"]
        A.remove_node(1)
        assert A.nodes() == []
        with pytest.raises(KeyError):
            A.remove_node(1)

    def test_add_remove_nodes_from(self):
        A = pgv.AGraph()
        A.add_nodes_from(range(3))
        assert sorted(A.nodes()) == ["0", "1", "2"]
        A.remove_nodes_from(range(3))
        assert A.nodes() == []
        with pytest.raises(KeyError):
            A.remove_nodes_from(range(3))

    def test_nodes(self):
        assert sorted(self.P3.nodes()) == ["1", "2", "3"]
        assert sorted(self.P3.nodes_iter()) == ["1", "2", "3"]
        assert sorted(self.P3.iternodes()) == ["1", "2", "3"]

    def test_order(self):
        assert self.P3.number_of_nodes() == 3
        assert self.P3.order() == 3

    def test_has_node(self):
        assert self.P3.has_node(1)
        assert self.P3.has_node("1")
        assert not self.P3.has_node("10")

    def test_get_node(self):
        assert self.P3.get_node(1) == "1"
        one = self.P3.get_node(1)
        nh = one.get_handle()
        node = pgv.Node(self.P3, 1)
        assert node.get_handle() == nh
        assert node == "1"
        with pytest.raises(KeyError):
            pgv.Node(self.P3, 10)

    def test_add_edge(self):
        A = pgv.AGraph()
        A.add_edge(1, 2)
        assert sorted(A.edges()) == [("1", "2")]
        e = (2, 3)
        A.add_edge(e)
        edges = [("1", "2"), ("2", "3")]
        assert sorted(tuple(sorted(e)) for e in A.edges()) == edges

    def test_add_remove_edges_from(self):
        A = pgv.AGraph()
        A.add_edges_from([(1, 2), (2, 3)])
        edges = [("1", "2"), ("2", "3")]
        assert sorted(tuple(sorted(e)) for e in A.edges()) == edges

        A.remove_edge(1, 2)
        assert sorted(A.edges()) == [("2", "3")]
        e = (2, 3)
        A.remove_edge(e)
        assert A.edges() == []
        with pytest.raises(KeyError):
            A.remove_edge(1, 2)

    def test_remove_edges_from(self):
        A = pgv.AGraph()
        A.add_edges_from([(1, 2), (2, 3)])
        edges = [("1", "2"), ("2", "3")]
        assert sorted(tuple(sorted(e)) for e in A.edges()) == edges
        A.remove_edges_from([(1, 2), (2, 3)])
        assert A.edges() == []

    def test_edges(self):
        A = pgv.AGraph()
        A.add_edges_from([(1, 2), (2, 3)])
        edges = [("1", "2"), ("2", "3")]
        assert sorted(tuple(sorted(e)) for e in A.edges()) == edges
        assert sorted(tuple(sorted(e)) for e in A.edges_iter()) == edges
        assert sorted(tuple(sorted(e)) for e in A.iteredges()) == edges

    def test_has_edge(self):
        assert self.P3.has_edge(1, 2)
        assert self.P3.has_edge("1", "2")
        assert not self.P3.has_edge("1", "10")
        assert self.P3.get_edge(1, 2) == ("1", "2")
        eone = self.P3.get_edge(1, 2)
        eh = eone.handle
        edge = pgv.Edge(self.P3, 1, 2)
        assert edge.handle == eh
        assert edge == ("1", "2")
        A = self.P3.copy()
        A.add_node(10)
        with pytest.raises(KeyError):
            pgv.Edge(A, 1, 10)

    def test_neighbors(self):
        A = pgv.AGraph()
        A.add_edges_from([(1, 2), (2, 3)])
        assert sorted(A.neighbors(2)) == ["1", "3"]
        assert sorted(A.neighbors_iter(2)) == ["1", "3"]
        assert sorted(A.iterneighbors(2)) == ["1", "3"]

    def test_degree(self):
        A = pgv.AGraph()
        A.add_edges_from([(1, 2), (2, 3)])
        assert sorted(A.degree()) == [1, 1, 2]
        assert sorted(A.degree_iter()) == [("1", 1), ("2", 2), ("3", 1)]
        assert sorted(A.iterdegree()) == [("1", 1), ("2", 2), ("3", 1)]
        assert sorted(A.iterdegree(A.nodes())) == [("1", 1), ("2", 2), ("3", 1)]
        assert sorted(A.iterdegree(A)) == [("1", 1), ("2", 2), ("3", 1)]
        assert A.degree(1) == 1
        assert A.degree(2) == 2

    def test_clear(self):
        A = pgv.AGraph()
        A.add_edges_from([(1, 2), (2, 3)])
        A.clear()
        assert A.nodes() == []
        assert A.edges() == []

    def test_copy(self):
        A = self.P3.copy()
        assert A.nodes() == ["1", "2", "3"]
        assert A.edges() == [("1", "2"), ("2", "3")]
        assert A is not self.P3
        assert self.P3 is self.P3
        assert stringify(A) == stringify(self.P3)
        assert A == self.P3

        # see Github Issue #354: G.copy() doesn't return a faithful copy
        DG = pgv.AGraph(directed=True)
        DG.add_edge(1, 2)
        DG_copy = DG.copy()
        assert DG_copy.is_directed()
        G = pgv.AGraph()
        G.add_edge(1, 2)
        G_copy = G.copy()
        assert not G_copy.is_directed()

        # Similarly with the strict and name attrs when copying: see gh-426
        A = pgv.AGraph(strict=False, directed=True, name="foobar")
        AC = A.copy()
        assert AC.strict == A.strict
        assert AC.name == A.name

    def test_multigraph_copy_with_keys(self):
        A = pgv.AGraph(strict=False)
        # Add parallel edges
        A.add_edge(1, 2, key="1_2-1")
        A.add_edge(1, 2, key="1_2-2")
        # Add single edge
        A.add_edge(3, 4, key="3_4-1")

        AC = A.copy()

        # Sanity - verify keys exist in original graph
        assert all(key is not None for key in A.edges(keys=True))
        # Verify all edges in copied graph have edges
        assert all(key is not None for key in AC.edges(keys=True))
        # Verify edges, including keys, are identical to original graph
        assert set(A.edges(keys=True)) == set(AC.edges(keys=True))

    def test_add_path(self):
        A = pgv.AGraph()
        A.add_path([1, 2, 3])
        assert A == self.P3

    def test_add_cycle(self):
        A = pgv.AGraph()
        A.add_cycle([1, 2, 3])
        assert A.nodes() == ["1", "2", "3"]
        edges = [("1", "2"), ("1", "3"), ("2", "3")]
        assert sorted(tuple(sorted(e)) for e in A.iteredges()) == edges

    def test_graph_strict(self):
        A = pgv.AGraph()
        A.add_node(1)
        A.add_node(1)  # silent falure
        assert A.nodes() == ["1"]
        A.add_edge(1, 2)
        A.add_edge(1, 2)  # silent falure
        assert A.edges() == [("1", "2")]
        A.add_edge(3, 3)  # self-loops OK with strict
        assert A.edges() == [("1", "2"), ("3", "3")]
        assert A.nodes() == ["1", "2", "3"]

    def test_graph_not_strict(self):
        A = pgv.AGraph(strict=False)
        assert not A.is_strict()
        A.add_node(1)
        A.add_node(1)  # nop
        assert A.nodes() == ["1"]
        A.add_edge(1, 2)
        A.add_edge(1, 2)
        assert A.edges() == [("1", "2"), ("1", "2")]
        A.add_edge(3, 3)
        assert A.edges() == [("1", "2"), ("1", "2"), ("3", "3")]
        assert A.nodes() == ["1", "2", "3"]

        A.add_edge("3", "3", "foo")
        assert A.edges() == [("1", "2"), ("1", "2"), ("3", "3"), ("3", "3")]
        ans = [("1", "2", None), ("1", "2", None), ("3", "3", None), ("3", "3", "foo")]
        assert A.edges(keys=True) == ans
        eone = A.get_edge(3, 3, "foo")
        eh = eone.handle
        edge = pgv.Edge(A, 3, 3, "foo")
        assert edge.handle == eh
        assert edge == ("3", "3")
        assert eone == ("3", "3")
        assert edge.name == "foo"


class TestDiGraphOnly(TestGraph):
    def test_edge(self):
        A = pgv.AGraph(directed=True)
        A.add_edge(1, 2)
        assert A.has_edge(1, 2)
        assert not A.has_edge(2, 1)
        A.add_edge(2, 1)
        assert A.has_edge(2, 1)

    def test_edges(self):
        A = pgv.AGraph(directed=True)
        A.add_edges_from(self.P3.edges())
        edges = [("1", "2"), ("2", "3")]
        assert sorted(tuple(sorted(e)) for e in A.edges()) == edges
        assert sorted(tuple(sorted(e)) for e in A.edges_iter()) == edges
        assert sorted(tuple(sorted(e)) for e in A.out_edges()) == edges
        assert sorted(tuple(sorted(e)) for e in A.out_edges_iter()) == edges
        assert sorted(tuple(sorted(e)) for e in A.in_edges()) == edges
        assert sorted(tuple(sorted(e)) for e in A.in_edges_iter()) == edges
        assert sorted(A.edges(1)) == [("1", "2")]
        assert sorted(A.edges([1, 2])) == [("1", "2"), ("2", "3")]
        assert sorted(A.edges_iter(1)) == [("1", "2")]
        assert sorted(A.out_edges(1)) == [("1", "2")]
        assert sorted(A.out_edges([1, 2])) == [("1", "2"), ("2", "3")]
        assert sorted(A.out_edges_iter(1)) == [("1", "2")]
        assert sorted(A.in_edges(2)) == [("1", "2")]
        assert sorted(A.in_edges([1, 2])) == [("1", "2")]
        assert sorted(A.in_edges_iter(2)) == [("1", "2")]
        assert A.predecessors(1) == []
        assert list(A.predecessors_iter(1)) == []
        assert A.predecessors(2) == ["1"]
        assert list(A.predecessors_iter(2)) == ["1"]
        assert A.successors(1) == ["2"]
        assert list(A.successors_iter(1)) == ["2"]
        assert A.successors(2) == ["3"]
        assert list(A.successors_iter(2)) == ["3"]
        assert sorted(A.out_degree()) == [0, 1, 1]
        assert sorted(A.out_degree(with_labels=True).values()) == [0, 1, 1]
        assert sorted(A.out_degree_iter()) == [("1", 1), ("2", 1), ("3", 0)]
        assert sorted(A.in_degree()) == [0, 1, 1]
        assert sorted(A.in_degree(with_labels=True).values()) == [0, 1, 1]
        assert sorted(A.in_degree_iter()) == [("1", 0), ("2", 1), ("3", 1)]

        assert A.in_degree(1) == 0
        assert A.out_degree(1) == 1
        assert A.in_degree(2) == 1
        assert A.out_degree(2) == 1

        ans = """strict digraph { 1 -> 2; 2 -> 3; }"""
        assert stringify(A) == ans

        ans = """strict digraph { 2 -> 1; 3 -> 2; }"""
        assert stringify(A.reverse()) == ans

    def test_name(self):
        A = pgv.AGraph(name="test")
        assert A.name == "test"
        B = A.reverse()
        assert B.name == "test"


@pytest.mark.parametrize(
    "d",
    (
        {0: {1: {}}, 1: {0: {}, 2: {}}, 2: {1: {}}},  # Dict-of-dicts
        {0: [1], 1: [0, 2], 2: [1]},  # Dict-of-lists
    ),
)
def test_agraph_constructor_dict_input(d):
    """Test AGraph constructor with thing = dod or dol."""
    # d is a dod or dol representation of a path graph with 3 nodes
    A = pgv.AGraph(d)
    assert sorted(A.nodes()) == ["0", "1", "2"]
    assert sorted(A.edges()) == [("0", "1"), ("1", "2")]


def test_agraph_constructor_handle_input():
    """Test AGraph constructor with thing= a Swig pointer - graph handle."""
    base = pgv.AGraph({0: [1], 1: [0, 2], 2: [1]})
    child = pgv.AGraph(base.handle)
    assert base == child
    # AGraphs created from handle reference original graph
    base.remove_node(2)
    assert sorted(child.nodes()) == ["0", "1"]


def test_agraph_constructor_bad_input():
    """AGraph constructor does not support edge lists."""
    with pytest.raises(TypeError, match="Unrecognized input"):
        pgv.AGraph([(0, 1), (1, 2)])


def test_agraph_constructor_string_non_standard_encoding():
    """AGraph constructor with a string specifying non-utf8-encoding."""
    # A path graph string in .dot format with charset specified
    dotstring = 'strict graph "" {\n\tcharset="latin1";\n\t0 -- 1;\n\t1 -- 2;\n}\n'
    A = pgv.AGraph(string=dotstring)
    assert A.encoding == "latin1"
    assert sorted(A.edges()) == [("0", "1"), ("1", "2")]


def test_agraph_equality_node_attrs():
    """Graphs are not equal if node attributes differ."""
    nodes = [0, 1]
    # Create graphs with the same nodes
    A, B = pgv.AGraph(), pgv.AGraph()
    A.add_nodes_from(nodes)
    B.add_nodes_from(nodes)
    # Set default attributes for all nodes in each graph
    A.node_attr["color"] = "red"
    B.node_attr["color"] = "red"
    assert A.get_node(1).attr["color"] == "red"
    assert A == B
    # Change attribute of a single node in B
    B.get_node(1).attr["color"] = "blue"
    # Graphs are no longer considered equal
    assert A != B
    # See #284  A and C should be different
    C = pgv.AGraph()
    A.add_nodes_from(nodes)
    assert A != C


def test_agraph_equality_edge_attrs():
    """Graphs are not equal if edge attributes differ."""
    A, B = pgv.AGraph(), pgv.AGraph()
    A.add_edge(0, 1, weight=1.0)
    B.add_edge(0, 1, weight=1.0)
    assert A == B
    # Change edge attribute
    B.get_edge(0, 1).attr["weight"] = 2.0
    assert A != B
    # See #284  A and C should be different
    C = pgv.AGraph()
    A.add_edge(0, 1)
    assert A != C


def test_agraph_has_edge_single_input_parsing():
    """If len(args) is 1, args[0] is assumed to be a tuple."""
    A = pgv.AGraph({0: [1], 1: [0, 2], 2: [1]})
    assert A.has_edge((0, 1))
    assert not A.has_edge((0, 3))


def test_repr_on_incomplete_initialization():
    """Smoke test to ensure no segfaults from accessing uninitialized attributes
    in __repr__ when object initialization fails. See gh-519."""
    with pytest.raises(TypeError, match="Unrecognized input"):
        A = pgv.AGraph(object())


@pytest.mark.parametrize(
    "node",
    (
        "Node D_0__Bacteria D_1__Cyanobacteria D_2__Chloroplast D_3__Bryum argenteum var. argenteum D_4__Bryum argenteum var. argenteum D_5__Bryum argenteum var. argenteum D_6__Bryum argenteum var",  # Very long node name
        "xxxxxx市中西医结合医疗健康集团xxxxx分院(xxx市xxxxxxxxxx镇卫生院、xxx市xxxxxxxxxxx镇社区卫生服务中心)(15094137051)",  # with unicode
    ),
)
def test_long_node(node):
    """Test agnode doesn't fail when graph contains very long nodes. See gh-123."""
    A = pgv.AGraph(directed=True)
    A.add_nodes_from(list(range(5)) + [node])
    A.layout()  # Smoke test - should not raise
    # Check each node was assigned a position
    assert A.to_string().count("pos") == 6



import warnings
import pygraphviz as pgv
import pytest


def test_drawing_hex_colors(tmp_path):
    """Ensure hex colors are supported without warnings - see gh-351."""
    # fmt: off
    dotstr = """
        digraph dag {
        ratio=fill
        node [style=filled,color="#444444",fillcolor="#ffed6f"]
        edge [arrowhead=normal,arrowsize=1.0]

        "ID0000001" [shape=ellipse,color="#000000",fillcolor="#1b9e77",label="preprocess\nID0000001"]
        "ID0000002" [shape=ellipse,color="#000000",fillcolor="#d95f02",label="findrange\nID0000002"]
        "ID0000003" [shape=ellipse,color="#000000",fillcolor="#d95f02",label="findrange\nID0000003"]
        "ID0000004" [shape=ellipse,color="#000000",fillcolor="#7570b3",label="analyze\nID0000004"]
        "f.a" [shape=rect,color="#000000",fillcolor="#ffed6f",label="f.a"]
        "f.b1" [shape=rect,color="#000000",fillcolor="#ffed6f",label="f.b1"]
        "f.b2" [shape=rect,color="#000000",fillcolor="#ffed6f",label="f.b2"]
        "f.c1" [shape=rect,color="#000000",fillcolor="#ffed6f",label="f.c1"]
        "f.c2" [shape=rect,color="#000000",fillcolor="#ffed6f",label="f.c2"]
        "f.d" [shape=rect,color="#000000",fillcolor="#ffed6f",label="f.d"]
        "ID0000001" -> "f.b1" [color="#000000"]
        "ID0000001" -> "f.b2" [color="#000000"]
        "ID0000001" -> "ID0000003" [color="#000000"]
        "ID0000001" -> "ID0000002" [color="#000000"]
        "ID0000002" -> "f.c1" [color="#000000"]
        "ID0000002" -> "ID0000004" [color="#000000"]
        "ID0000003" -> "f.c2" [color="#000000"]
        "ID0000003" -> "ID0000004" [color="#000000"]
        "ID0000004" -> "f.d" [color="#000000"]
        "f.a" -> "ID0000001" [color="#000000"]
        "f.b1" -> "ID0000002" [color="#000000"]
        "f.b2" -> "ID0000003" [color="#000000"]
        "f.c1" -> "ID0000004" [color="#000000"]
        "f.c2" -> "ID0000004" [color="#000000"]
    }
    """
    # fmt: on
    A = pgv.AGraph()
    A.from_string(dotstr)
    with warnings.catch_warnings(record=True) as rec:
        A.layout("dot")
        A.draw(str(tmp_path / "img.png"))
    assert len(rec) == 0


def test_drawing_error():
    with pytest.raises(AttributeError):
        A = pgv.AGraph(name="test graph")
        A.add_path([1, 2, 3, 4])
        d = A.draw()


def test_name_error():
    with pytest.raises(ValueError):
        A = pgv.AGraph(name="test graph")
        A.draw("foo", prog="foo")


def test_drawing_no_error_with_no_layout():
    A = pgv.AGraph(name="test graph")
    A.add_path([1, 2, 3, 4])
    d = A.draw(prog="nop")
    A.string_nop()


def test_drawing_makes_file_handle(tmp_path):
    A = pgv.AGraph(name="test graph")
    A.add_path([1, 2, 3, 4])
    with (tmp_path / "file1").open("wb") as fh:
        A.draw(fh, format="png", prog="twopi")
        assert fh.tell() > 0
    with (tmp_path / "file2").open("wb") as fh:
        A.draw(path=fh, prog="circo", format="png")
        assert fh.tell() > 0


def test_drawing_makes_file_path(tmp_path):
    A = pgv.AGraph(name="test graph")
    A.add_path([1, 2, 3, 4])
    out = tmp_path / "draw.png"
    A.draw(out, format="png", prog="twopi")
    assert out.is_file()
    out2 = tmp_path / "draw.png"
    A.draw(str(out2), format="png", prog="twopi")
    assert out2.is_file()


def test_drawing_to_create_dot_string():
    A = pgv.AGraph(name="test graph")
    A.add_path([1, 2, 3, 4])
    A.layout()
    dot_rep = A.to_string()
    assert "test graph" in dot_rep
    assert "strict graph" in dot_rep
    assert "pos" in dot_rep
    assert "height" in dot_rep
    assert "width" in dot_rep
    assert "1 -- 2" in dot_rep
    assert "2 -- 3" in dot_rep
    assert "3 -- 4" in dot_rep

    # unfortunately, the layout and dot outcomes vary
    # with system and graphviz version. One example is
    # shown here, the numbers can be very different.
    expected = """strict graph "test graph" {
	graph [bb="0,0,70.071,250.3"];
	node [label="\\N"];
	1	[height=0.5,
		pos="27,18",
		width=0.75];
	2	[height=0.5,
		pos="43.071,88.469",
		width=0.75];
	1 -- 2	[pos="31.139,36.148 33.557,46.75 36.596,60.077 39.002,70.627"];
	3	[height=0.5,
		pos="41.467,160.69",
		width=0.75];
	2 -- 3	[pos="42.666,106.69 42.423,117.64 42.115,131.52 41.872,142.47"];
	4	[height=0.5,
		pos="32.966,232.3",
		width=0.75];
	3 -- 4	[pos="39.322,178.76 38.043,189.53 36.424,203.17 35.14,213.98"];
}
"""
    # print("dot representation:", dot_rep)
    # assert expected == dot_rep


class TestExperimentalGraphvizLibInterface:
    def test_drawing_error(self):
        with pytest.raises(AttributeError):
            A = pgv.AGraph(name="test graph")
            A.add_path([1, 2, 3, 4])
            d = A.draw()

    def test_name_error(self):
        with pytest.raises(ValueError):
            A = pgv.AGraph(name="test graph")
            A.draw("foo", prog="foo")

    def test_drawing_no_error_with_no_layout(self):
        A = pgv.AGraph(name="test graph")
        A.add_path([1, 2, 3, 4])
        d = A.draw(prog="nop")
        A.string_nop()

    def test_drawing_png_output_with_NULL_smoketest(self):
        """The PNG format can contain NULL bytes, which can cause cstring
        allocation problems if the size isn't handled correctly. See gh-424.
        """
        A = pgv.AGraph(name="test graph")
        A.add_path([1, 2, 3, 4])
        d = A.draw(prog="dot", format="png")

    def test_drawing_to_create_dot_string(self):
        A = pgv.AGraph(name="test graph")
        A.add_path([1, 2, 3, 4])
        A.layout()
        dot_rep = A.to_string()
        assert "test graph" in dot_rep
        assert "strict graph" in dot_rep
        assert "pos" in dot_rep
        assert "height" in dot_rep
        assert "width" in dot_rep
        assert "1 -- 2" in dot_rep
        assert "2 -- 3" in dot_rep
        assert "3 -- 4" in dot_rep

        # unfortunately, the layout and dot outcomes vary
        # with system and graphviz version. One example is
        # shown here, the numbers can be very different.
        expected = """strict graph "test graph" {
        graph [bb="0,0,70.071,250.3"];
        node [label="\\N"];
        1	[height=0.5,
            pos="27,18",
            width=0.75];
        2	[height=0.5,
            pos="43.071,88.469",
            width=0.75];
        1 -- 2	[pos="31.139,36.148 33.557,46.75 36.596,60.077 39.002,70.627"];
        3	[height=0.5,
            pos="41.467,160.69",
            width=0.75];
        2 -- 3	[pos="42.666,106.69 42.423,117.64 42.115,131.52 41.872,142.47"];
        4	[height=0.5,
            pos="32.966,232.3",
            width=0.75];
        3 -- 4	[pos="39.322,178.76 38.043,189.53 36.424,203.17 35.14,213.98"];
    }
    """
        # print("dot representation:", dot_rep)
        # assert expected == dot_rep

    def test_drawing_makes_file(self, tmp_path):
        A = pgv.AGraph(name="test graph")
        A.add_path([1, 2, 3, 4])
        with (tmp_path / "file").open("wb") as fh:
            A.draw(fh, format="png", prog="twopi")
            assert fh.tell() > 0

    def test_drawing_makes_file1(self, tmp_path):
        A = pgv.AGraph(name="test graph")
        A.add_path([1, 2, 3, 4])
        with (tmp_path / "file").open("wb") as fh:
            A.draw(path=fh, prog="circo", format="png")
            assert fh.tell() > 0



import pytest
import pygraphviz as pgv

graphviz_major_version = int(pgv.__graphviz_version__.split(".")[0])


@pytest.mark.skipif(graphviz_major_version < 13, reason="Graphviz version too old")
def test_multiple_reads_same_source_trailing_character(tmp_path):
    """Ensure multiple reads from the same text file with an unexpected trailing
    character don't cause agread fails. See gh-171."""
    fpath = tmp_path / "hello.gv.txt"
    # Original bug was in the read-from-file code path
    with open(fpath, "w") as fh:
        fh.write('digraph G {Hello->World}"')  # note the trailing "

    # Smoke test: sequential reads from the same file via the AGraph constructor
    # should not raise ValueError/DotError
    A = pgv.AGraph(str(fpath))
    B = pgv.AGraph(str(fpath))
    assert A.to_string() == B.to_string()


def test_readwrite(tmp_path):
    A = pgv.AGraph(name="test graph")
    A.add_path([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    path = tmp_path / "rw_win32_testfile"
    with path.open("w") as fh:
        assert not fh.closed
    # Make sure it can be opened for writing again on Win32
    assert fh.closed
    # Pass a string to trigger the code paths that close the newly created file handle
    A.write(path)
    B = pgv.AGraph(path)
    assert A == B
    assert B == A
    assert B is not A


def test_readwrite_pathobj(tmp_path):
    A = pgv.AGraph(name="test graph")
    A.add_path([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    path = tmp_path / "rw_win32_testfile"
    with path.open("w") as fh:
        assert not fh.closed
    # Make sure it can be opened for writing again on Win32
    assert fh.closed
    # Pass a string to trigger the code paths that close the newly created file handle
    A.write(path)
    B = pgv.AGraph(path)
    assert A == B
    assert B == A
    assert B is not A


def test_sequential_reads_windows():
    for _ in range(512):
        pgv.AGraph("digraph {1 -> 2}")



import pytest

import pygraphviz as pgv


def test_sequential_layout_filehandles_windows():
    """See gh-483"""
    # Check that a bunch of sequential layout calls doesn't bork windows
    for _ in range(512):
        A = pgv.AGraph()
        A.layout()
    # The existence of "node" in the graph is an indication that the layout
    # was run successfully.
    assert "node" in A.to_string()


def test_layout():
    A = pgv.AGraph(name="test graph")
    A.add_path([1, 2, 3, 4])
    assert [n.attr["pos"] is None for n in A.nodes()] == [True] * 4
    A.layout()
    assert [n.attr["pos"] is not None for n in A.nodes()] == [True] * 4


def test_layout_defaults():
    A = pgv.AGraph(name="test graph")
    A.add_path([1, 2, 3, 4])
    # print("Pos before",[n.attr["pos"] for n in A.nodes()])
    assert [n.attr["pos"] is None for n in A.nodes()] == [True] * 4
    A.layout()
    assert [n.attr["pos"] is not None for n in A.nodes()] == [True] * 4
    # print("Pos after",[n.attr["pos"] for n in A.nodes()])


@pytest.mark.parametrize(
    "prog",
    ("neato", "dot", "twopi", "circo", "fdp", "osage", "patchwork", "sfdp"),
)
def test_layout_prog_arg(prog):
    A = pgv.AGraph(name="test graph")
    A.add_path([1, 2, 3, 4])

    # No node position before layout
    assert all(n.attr["pos"] is None for n in A.nodes())

    # All nodes should have a position after layout
    A.layout(prog=prog)
    assert all(n.attr["pos"] is not None for n in A.nodes())

    # Node positions should be different for each of the layouts
    pos = [n.attr["pos"] for n in A.nodes()]
    A.layout(prog="dot")  # Use dot layout as reference
    dot_pos = [n.attr["pos"] for n in A.nodes()]
    assert pos == dot_pos if prog == "dot" else pos != dot_pos


def test_bad_prog_arg_raises():
    A = pgv.AGraph()
    A.add_path([1, 2, 3, 4])
    with pytest.raises(ValueError, match="Program.*is not a valid layout"):
        A.layout(prog="not-a-valid-layout")


class TestExperimentalGraphvizLibInterface:
    def test_layout(self):
        A = pgv.AGraph(name="test graph")
        A.add_path([1, 2, 3, 4])
        assert [n.attr["pos"] is None for n in A.nodes()] == [True] * 4
        A.layout()
        assert [n.attr["pos"] is not None for n in A.nodes()] == [True] * 4

    def test_layout_defaults(self):
        A = pgv.AGraph(name="test graph")
        A.add_path([1, 2, 3, 4])
        # print("Pos before",[n.attr["pos"] for n in A.nodes()])
        assert [n.attr["pos"] is None for n in A.nodes()] == [True] * 4
        A.layout()
        assert [n.attr["pos"] is not None for n in A.nodes()] == [True] * 4
        # print("Pos after",[n.attr["pos"] for n in A.nodes()])

    def test_layout_prog_arg(self):
        A = pgv.AGraph(name="test graph")
        A.add_path([1, 2, 3, 4])
        assert [n.attr["pos"] is None for n in A.nodes()] == [True] * 4
        A.layout(prog=b"dot")
        assert [n.attr["pos"] is not None for n in A.nodes()] == [True] * 4
        dot_pos = [n.attr["pos"] for n in A.nodes()]

        A.layout(prog="dot")
        result = [n.attr["pos"] for n in A.nodes()]
        assert result == dot_pos

        A.layout(prog="twopi")
        result = [n.attr["pos"] for n in A.nodes()]
        assert result != dot_pos

        A.layout(prog="neato")
        result = [n.attr["pos"] for n in A.nodes()]
        assert result != dot_pos

        A.layout(prog="circo")
        result = [n.attr["pos"] for n in A.nodes()]
        assert result != dot_pos

        A.layout(prog="fdp")
        result = [n.attr["pos"] for n in A.nodes()]
        assert result != dot_pos

        A.layout(prog="nop")
        result = [n.attr["pos"] for n in A.nodes()]
        assert result != dot_pos



"""
PyGraphviz
==========

A Python wrapper for the graphviz Agraph data structure.

See https://pygraphviz.github.io for complete documentation.
See pygraphviz.AGraph for detailed documentation.
"""

import sys

# https://docs.python.org/3/whatsnew/3.8.html#bpo-36085-whatsnew
if sys.version_info >= (3, 8, 0) and sys.platform == "win32":
    import os

    # Wheel install: delvewheel places DLLs in pygraphviz.libs/
    _libs_dir = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "pygraphviz.libs"
    )
    if os.path.isdir(_libs_dir):
        os.add_dll_directory(_libs_dir)
    # Dev install: find graphviz on PATH
    for path in os.environ.get("PATH", "").split(os.pathsep):
        if "graphviz" in path.lower() and os.path.exists(path):
            os.add_dll_directory(path)


from ._graphviz import (
    GRAPHVIZ_MAJOR_VERSION,
    GRAPHVIZ_MINOR_VERSION,
    GRAPHVIZ_PATCH_VERSION,
)

__version__ = "2.0.1"
__graphviz_version__ = (
    f"{GRAPHVIZ_MAJOR_VERSION}.{GRAPHVIZ_MINOR_VERSION}.{GRAPHVIZ_PATCH_VERSION}"
)

from .agraph import AGraph, Attribute, DotError, Edge, ItemAttribute, Node

__all__ = ["AGraph", "Attribute", "DotError", "Edge", "ItemAttribute", "Node"]

from . import testing

# Per contract with Sphinx-Gallery, this method must be available at top level
from pygraphviz.scraper import _get_sg_image_scraper

del sys
